"""
Copyright 2023 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

from asyncpg import Connection


async def update(connection: Connection) -> None:
    schema = """
    CREATE TYPE auth_method AS ENUM('database', 'oidc');

    CREATE TABLE IF NOT EXISTS public.user(
        id uuid PRIMARY KEY,
        username VARCHAR NOT NULL UNIQUE ,
        password_hash VARCHAR NOT NULL,
        auth_method auth_method NOT NULL
    );
    """

    await connection.execute(schema)
